require 'criteria'

module Criteria
  class SQLTable < Table
    attr_reader :table

    def initialize(table)
      @table          = table
      @criterion_type = SQLCriterion
    end

    def to_s
      return @table
    end
  end #c:SQLTable


  class SQLCriterion < Criterion
    attr_accessor :limit, :offset, :order, :order_by

    protected
    attr_accessor :join_on, :joins, :join_type

    def _mkselect(s, *args)
      # SELECT and column clause
      s << "SELECT "
      s << args.map {
        |s|
        if s.is_a?(SQLTable); "#{s}.*"
        else                  s.to_s
        end
      }.join(", ")

      s << "*" unless args.size > 0
    end #m:_mkselect

    def _mkfrom(s, *args)
      # FROM clause
      s << " FROM "
      s << "#{@_parent.table}"    if(@_parent.table && args.size == 0)
      s << args.map { |i| i.is_a?(SQLCriterion) ? i._parent.table : i }.join(", ")
    end

    def _mkjoin(s, *args)
      return unless @join_type

      # JOIN clause
      s << " #@join_type "
      s << @joins.map { |j| j.to_s }.join(", ")

      if(@join_on.is_a? SQLCriterion)
        s << " ON "
        s << @join_on.to_s
      elsif(@join_on.is_a? Array)
        s << " USING("
        s << @join_on.join(", ")
        s << ")"
      end
    end #m:_mkjoin

    def _mkwhere(s, *args)
      # WHERE clause
      s << " WHERE #{self.to_s}"
    end #m_mkwhere

    def _mklimit(s, *args)
      # LIMIT and OFFSET clause
      if @limit
        s << " LIMIT"
        s << " #@offset, "   if @offset
        s << " #@limit"
      end
    end #m:_mklimit

    def _mkorder(s, *args)
      # ORDER clause
      if @order_by
        s << " ORDER BY #{@order_by.join(', ')}"
        s << " #@order"                             if order
      end
    end #m:_mkorder

    def self._defjoin(name, type, on = false, *args)
      dupstr = "q = self.dup\n"

      if on
        s  = "def #{name.to_s}(on, *args)\n"
        s << dupstr
        s << "q.join_on = on\n"
      else
        s  = "def #{name.to_s}(*args)\n"
        s << dupstr
      end

      s << <<-END
           q.joins     = args
           q.join_type = "#{type}"
           return q
        end
      END

      self.class_eval s
    end
    public

    def select(*args)
      s = ""
      _mkselect(s, *args)
      _mkfrom(s, *args)
      _mkjoin(s, *args)
      _mkwhere(s, *args)
      _mklimit(s, *args)
      _mkorder(s, *args)
      return s
    end #m:select

    _defjoin(:join,        "JOIN")
    _defjoin(:left_join,   "LEFT JOIN")
    _defjoin(:right_join,  "RIGHT JOIN")
    _defjoin(:full_join,   "FULL JOIN")
    _defjoin(:cross_join,  "CROSS JOIN")

    _defjoin(:inner_join,             "INNER JOIN",         true)
    _defjoin(:left_outer_join,        "LEFT OUTER JOIN",    true)
    _defjoin(:right_outer_join,       "RIGHT OUTER JOIN",   true)
    _defjoin(:full_outer_join,        "FULL OUTER JOIN",    true)

    def to_s
      if @chain.size < 1
        s  = ""
        s << "#{@_parent.table}."     if(@_parent.table && @field.is_a?(Symbol))
        s << @field.to_s
        return s
      end

      s  = "("
      s << "#{@_parent.table}."       if(@_parent.table && @field.is_a?(Symbol))
      s << "#{@field.to_s} #{_translate_fn @chain[0]}"
      @chain[1...@chain.size].each {
        | item |
        s << " "
        s << "#{@_parent.table}."     if(@_parent.table && item.is_a?(Symbol))
        s << _translate_item(item)
      }

      s << ")"
      return s
    end #m:to_s

    def _translate_fn(op)
      case op
      when :&;  "AND"
      when :|;  "OR"
      when :==; "="
      else;     op
      end
    end #m:_translate_fn

    def _translate_item(item)
      case item
      when String; "'#{item}'"
      when Regexp; "'#{item}'"
      else;        item.to_s
      end
    end #m:_translate_item
    protected :_translate_fn, :_translate_item

  end #c:SQLCriterion
end #M:Criteria
